#ifndef PEBBLISP_PLFUNC_H
#define PEBBLISP_PLFUNC_H

#include "pebblisp.h"

fn(add, "+",
   "Add numbers",
   "(+ 1 2)", "3",
   "(+ 1 2 3 4)", "10",
);

fn(sub, "-",
   "Subtract numbers",
   "(- 3 2)", "1",
   "(- 10 3 1)", "6",
);

fn(mul, "*",
   "Multiply numbers",
   "(* 3 2)", "6",
   "(* 1 2 3 4)", "24",
);

fn(dvi, "/",
   "Divide numbers",
   "(/ 8 2)", "4",
   "(/ 100 5 2)", "10",
);

fn(mod, "%",
   "Get the modulus of two numbers",
   "(% 10 3)", "1",
);

fn(equ, "=",
   "Check if two values are equal",
   "(= (+ 2 2) 4)", "T",
   "(= \"Hello\" (cat \"He\" \"llo\"))", "T",
   "(= (1 2 3) (1 2 3))", "T",
   "(= 10 \"10\")", "F",
   "(= 2 5)", "F",
);

fn(greaterThan, ">",
   "Check if one number is greater than another",
   "(> 10 3)", "T",
   "(> 2 5)", "F",
);

fn(lessThan, "<",
   "Check if one number is less than another",
   "(< 2 5)", "T",
   "(< 10 3)", "F",
);

fn(and, "&",
   "Check if all conditions are true",
   "(& T T)", "T",
   "(& F T)", "F",
   "(& T T T T T F)", "F",
);

fn(or, "|",
   "Check if any conditions are true",
   "(| T F)", "T",
   "(| F F)", "F",
   "(| F F F F F T)", "T",
);

tfn(catObjects, "cat",
    ({ anyType, returns(isStringy) }),
    "Concatenate string representations of the given objects.",
    "(cat \"Stuff: \" (1 2 3))", "Stuff: ( 1 2 3 )",
);

tfn(filter, "fil",
    ({ expect(isFuncy), expect(isListy), returns(isListy) }),
    "Filter a list based on the given condition.",
    "(fil (fn (a) (< 50 a)) (25 60 100))", "( 60 100 )",
    "(fil (fn (a) (< 0 (len a))) ( () (1) (1 2) () ))", "( ( 1 ) ( 1 2 ) )",
);

tfn(append, "ap",
    ({ expect(isListy), anyType, returns(isListy) }),
    "Append the given element. Creates a new list.",
    "(ap (1 2) 3)", "( 1 2 3 )",
);

tfn(prepend, "pre",
    ({ expect(isListy), anyType, returns(isListy) }),
    "Prepend the given element. Creates a new list",
    "(pre (2 3) 1)", "( 1 2 3 )",
);

tfn(len, "len",
    ({ expect(isListy), returns(isNumber) }),
    "Returns the length of the given list, or an error if the expression is not a list.",
    "(len (2 3))", "2",
    "(len ())", "0",
    "(len \"string\")", "BAD_PARAMS;",
);

tfn(reduce, "reduce",
    ({ anyType, expect(isFuncy), anyType, anyType }),
    "Performs a simple reduction.\n"
    "Takes three arguments:\n"
    " - Values\n"
    " - A function to apply to each value\n"
    " - An initial value",
    "(reduce 5 + 6)", "11",
    "(reduce (1 2 3) + 0)", "6",
    "(def sum (fn (left right) ((+ left right)) )) (reduce 5 sum 6)", "11",
    "(def sum (fn (left right) ((+ left right)) )) (reduce (1 2 3) sum 0)", "6",
);

tfn(at, "at",
    ({ expect(isNumber), expect(isListy), anyType }),
    "Get item at the given index in the given list.",
    "(at 1 (1 2 3))", "2",
    "(at 99 (1 2 3))", "INDEX_PAST_END",
    "(at 99 \"string\")", "BAD_PARAMS;",
);

tfn(rest, "rest",
    ({ expect(isListy), returns(isListy) }),
    "Get the tail of a list. All but the first element.",
    "(rest (1 2 3))", "( 2 3 )",
    "(rest ())", "( )",
    "(rest \"string\")", "BAD_PARAMS;",
);

tfn(reverse, "rev",
    ({ expect(isListy), returns(isListy) }),
    "Reverse a list.",
    "(rev (1 2 3))", "( 3 2 1 )",
    "(rev \"string\")", "BAD_PARAMS;",
);

tfn(isNum, "isnum",
    ({ anyType, returns(isBool) }),
    "Returns `T` only if the argument evaluates to a number.",
    "(isnum 1)", "T",
    "(isnum (+ 5 5))", "T",
    "(isnum '(+ 5 5))", "F",
    "(isnum \"Hello\")", "F",
);

tfn(isList, "islist",
    ({ anyType, returns(isBool) }),
    "Returns `T` only if the argument is a list.",
    "(islist (1 2 3))", "T",
    "(islist ())", "T",
    "(islist \"Stringy\")", "F",
);

tfn(isString, "isstr",
    ({ anyType, returns(isBool) }),
    "Returns `T` only if the argument is a string.",
    "(isstr \"Heyo\")", "T",
    "(isstr \"\")", "T",
    "(isstr (cat 5 5))", "T",
    "(isstr 10)", "F",
);

tfn(isErr, "iserr",
    ({ anyType, returns(isBool) }),
    "Check if the argument is an error.",
    "(iserr (at 10 ()))", "T",
    "(iserr 5)", "F",
);

tfn(charAt, "chat",
    ({ expect(isStringy), expect(isNumber), returns(isStringy) }),
    "Get the char in the given string at the given index.",
    "(chat \"Hello\" 1)", "e",
    "(chat \"Hello\" 10)", "",
);

tfn(charVal, "char",
    ({ expect(isStringy), returns(isNumber) }),
    "Get the ascii integer representaton of the given character.",
    "(char \"h\")", "104",
    "(char \"hello\")", "104",
    "(char \"\")", "0",
);

tfn(slen, "slen",
    ({ expect(isStringy), returns(isNumber) }),
    "Returns the length of the given string",
    "(slen \"string\")", "6",
    "(slen \"\")", "0",
);

tfn(chars, "chars",
    ({ expect(isStringy), returns(isListy) }),
    "Get a list of all chars in the given string",
    "(chars \"hello\")", "( h e l l o )",
    "(chars \"\")", "( )",
);

tfn(matches, "matches",
    ({ expect(isStringy), expect(isStringy), returns(isBool) }),
    "Check that a string matches a basic wildcard pattern\n"
    "Note: Currently there is no way to match a literal asterisk",
    "(matches \"Hiya\" \"Hiya\")", "T",
    "(matches \"Howdy\" \"H*y\")", "T",
    "(matches \"Yello\" \"*\")", "T",
    "(matches \"S\" \"*S\")", "T",
    "(matches \"S\" \"S*\")", "T",
    "(matches \"\" \"*\")", "T",
    "(matches \"ThisIsALongerOne\" \"This*ALong*One\")", "T",
    "(matches \"ThisIsALongerOne\" \"*This*ALong*One*\")", "T",
    ";", "",
    "(matches \"Howdy\" \"*llo\")", "F",
    "(matches \"Stink\" \"Stank\")", "F",
);

tfn(substring, "substr",
    ({ expect(isNumber), expect(isNumber), expect(isStringy), returns(isStringy) }),
    "Get a substring from the given string.",
    "(substr 1 3 \"Hello\")", "el",
    "(substr 99 3 \"Hello\")", "BAD_PARAMS;",
    "(substr 98 99 \"Hello\")", "BAD_PARAMS;",
    "(substr 0 1 \"\")", "",
    "(substr 0 99 \"Hey\")", "Hey",
    "(substr 1 99 \"Heyyyy\")", "eyyyy",
);

/// STRING/SLIST => ANY
fn(parseEvalO, "eval",
   "Evaluate the given string or quoted list. Uses the global scope.",
   "(eval \"(1 2 3)\")", "( 1 2 3 )",
   "(eval '(+ 5 5))", "10",
);

tfn(getTime, "time",
    ({ returns(isStruct) }),
    "Get a struct of the current time with fields (minute hour sec)."
);

#ifdef STANDALONE

fn(print, "prn", "Prints the string representation of the given object to stdout.");

tfn(numToChar, "ch",
    ({ expect(isNumber), returns(isStringy) }),
    "Gets a string containing the ascii character for the given number value.",
    "(ch 107)", "k",
    "(ch 0x21)", "!",
);

fn(printEnvO, "penv",
   "Prints out the current scoped environment.\n"
   "(penv) prints a mostly human-readable list of env variables.\n"
   "(penv T) prints a mostly list of env variables, including pointer addresses.\n"
   "Calling (penv) with no argument is equivalent to (penv F)"
);

tfn(systemCall, "sys",
    ({ expect(isStringy), returns(isNumber) }),
    "Opens a shell and runs the given command, returning 0 if successful.\n"
    "If the argument is not a string, returns 255.\n",
    "(sys \"echo yee > /dev/null\")", "0",
);

tfn(cd, "cd",
    ({ expect(isStringy), anyType }),
    "Change the current directory.",
    "(cd \"/\") (cwd)", "/"
);

tfn(cwd, "cwd",
    ({ returns(isStringy) }),
    "Get the current directory.",
    "(cd \"/\") (cwd)", "/"
);

/// @code
/// () => STRING
/// STRING => STRING
fn(takeInput, "inp",
   "Take console input with an optional prompt. For example:\n"
   "`(def x (input))` will wait for user input with no prompt.\n"
   "`(def x (input \">> \"))` wait for input, but prompt the user with '>> '\n"
);

tfn(readFileToObject, "rf",
    ({ expect(isStringy), returns(isStringy) }),
    "Read a file into a string object."
);

tfn(getEnvVar, "env",
    ({ expect(isStringy), returns(isStringy) }),
    "Get a variable from the current environment\n"
    "(env HOME) => /home/sagevaillancourt"
);

#endif // STANDALONE

#endif // PEBBLISP_PLFUNC_H
